@@ -1,12 +1,14 @@
 import logging
 from typing import Iterator, Optional
 
-from . import VEXObject
-from archinfo import RegisterOffset, TmpVar
-from .enums import get_enum_from_int, get_int_from_enum
-
-l = logging.getLogger('pyvex.stmt')
+from . import expr
+from .const import IRConst
+from .enums import IRCallee, IRRegArray, VEXObject, get_enum_from_int, get_int_from_enum
+from .errors import PyVEXError
+from .expr import Const, Get, IRExpr
+from .native import ffi, pvc
 
+log = logging.getLogger("pyvex.stmt")
 
 
 class IRStmt(VEXObject):
@@ -14,24 +16,28 @@ class IRStmt(VEXObject):
     IR statements in VEX represents operations with side-effects.
     """
 
-    tag = None # type: Optional[str]
+    tag: Optional[str] = None
     tag_int = 0  # set automatically at bottom of file
 
-    __slots__ = [ ]
+    __slots__ = []
 
     def pp(self):
-        print(self.__str__())
+        print(str(self))
 
     @property
-    def expressions(self) -> Iterator['IRExpr']:
+    def child_expressions(self) -> Iterator["IRExpr"]:
         for k in self.__slots__:
             v = getattr(self, k)
             if isinstance(v, IRExpr):
                 # return itself
                 yield v
                 # return all the child expressions
-                for child in v.child_expressions:
-                    yield child
+                yield from v.child_expressions
+
+    # ???
+    @property
+    def expressions(self):
+        return self.child_expressions
 
     @property
     def constants(self):
@@ -45,41 +51,43 @@ class IRStmt(VEXObject):
         try:
             stmt_class = enum_to_stmt_class(c_stmt.tag)._from_c(c_stmt)
         except KeyError:
-            raise PyVEXError('Unknown/unsupported IRStmtTag %s.\n' % get_enum_from_int(c_stmt.tag))
+            raise PyVEXError("Unknown/unsupported IRStmtTag %s.\n" % get_enum_from_int(c_stmt.tag))
         return stmt_class._from_c(c_stmt)
 
-    def typecheck(self, tyenv): # pylint: disable=unused-argument,no-self-use
+    def typecheck(self, tyenv):  # pylint: disable=unused-argument,no-self-use
         return True
 
-    def replace_expression(self, expression, replacement):
+    def replace_expression(self, replacements):
         """
         Replace child expressions in-place.
 
-        :param IRExpr expression:         The expression to look for.
-        :param IRExpr replacement:  The expression to replace with.
+        :param Dict[IRExpr, IRExpr] replacements:  A mapping from expression-to-find to expression-to-replace-with
         :return:                    None
         """
 
         for k in self.__slots__:
             v = getattr(self, k)
-            if v is expression:
-                setattr(self, k, replacement)
+            if isinstance(v, IRExpr) and v in replacements:
+                setattr(self, k, replacements.get(v))
             elif isinstance(v, IRExpr):
-                v.replace_expression(expression, replacement)
+                v.replace_expression(replacements)
             elif type(v) is tuple:
                 # Rebuild the tuple
-                _lst = [ ]
+                _lst = []
                 replaced = False
                 for expr_ in v:
-                    if expr_ is expression:
-                        _lst.append(replacement)
+                    if isinstance(expr_, IRExpr) and expr_ in replacements:
+                        _lst.append(replacements.get(expr_))
                         replaced = True
                     else:
                         _lst.append(expr_)
                 if replaced:
                     setattr(self, k, tuple(_lst))
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def __str__(self):
+        return self.pp_str(None, None, None)
+
+    def pp_str(self, reg_name=None, arch=None, tyenv=None) -> str:
         raise NotImplementedError()
 
 
@@ -88,11 +96,11 @@ class NoOp(IRStmt):
     A no-operation statement. It is usually the result of an IR optimization.
     """
 
-    __slots__ = [ ]
+    __slots__ = []
 
-    tag = 'Ist_NoOp'
+    tag = "Ist_NoOp"
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "IR-NoOp"
 
     @staticmethod
@@ -107,23 +115,21 @@ class IMark(IRStmt):
     instruction.
     """
 
-    __slots__ = ['addr', 'len', 'delta']
+    __slots__ = ["addr", "len", "delta"]
 
-    tag = 'Ist_IMark'
+    tag = "Ist_IMark"
 
     def __init__(self, addr: int, length: int, delta: int):
         self.addr = addr
         self.len = length
         self.delta = delta
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "------ IMark(0x%x, %d, %d) ------" % (self.addr, self.len, self.delta)
 
     @staticmethod
     def _from_c(c_stmt):
-        return IMark(c_stmt.Ist.IMark.addr,
-                     c_stmt.Ist.IMark.len,
-                     c_stmt.Ist.IMark.delta)
+        return IMark(c_stmt.Ist.IMark.addr, c_stmt.Ist.IMark.len, c_stmt.Ist.IMark.delta)
 
 
 class AbiHint(IRStmt):
@@ -131,51 +137,51 @@ class AbiHint(IRStmt):
     An ABI hint, provides specific information about this platform's ABI.
     """
 
-    __slots__ = ['base', 'len', 'nia']
+    __slots__ = ["base", "len", "nia"]
 
-    tag = 'Ist_AbiHint'
+    tag = "Ist_AbiHint"
 
     def __init__(self, base, length, nia):
         self.base = base
         self.len = length
         self.nia = nia
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "====== AbiHint(0x%s, %d, %s) ======" % (self.base, self.len, self.nia)
 
     @staticmethod
     def _from_c(c_stmt):
-        return AbiHint(IRExpr._from_c(c_stmt.Ist.AbiHint.base),
-                       c_stmt.Ist.AbiHint.len,
-                       IRExpr._from_c(c_stmt.Ist.AbiHint.nia))
+        return AbiHint(
+            IRExpr._from_c(c_stmt.Ist.AbiHint.base), c_stmt.Ist.AbiHint.len, IRExpr._from_c(c_stmt.Ist.AbiHint.nia)
+        )
+
 
 class Put(IRStmt):
     """
     Write to a guest register, at a fixed offset in the guest state.
     """
 
-    __slots__ = ['data', 'offset']
+    __slots__ = ["data", "offset"]
 
-    tag = 'Ist_Put'
+    tag = "Ist_Put"
 
-    def __init__(self, data: 'IRExpr', offset: RegisterOffset):
+    def __init__(self, data: "IRExpr", offset):
         self.data = data
         self.offset = offset
 
     ## TODO: Check if result_size and arch are available before looking of arch register name
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         if arch is not None and tyenv is not None:
             reg_name = arch.translate_register_name(self.offset, self.data.result_size(tyenv) // 8)
 
         if reg_name is not None:
-            return "PUT(%s) = %s" % (reg_name, self.data)
+            return f"PUT({reg_name}) = {self.data}"
         else:
-            return "PUT(offset=%s) = %s" % (self.offset, self.data)
+            return f"PUT(offset={self.offset}) = {self.data}"
 
     @staticmethod
     def _from_c(c_stmt):
-        return Put(IRExpr._from_c(c_stmt.Ist.Put.data),
-                   c_stmt.Ist.Put.offset)
+        return Put(IRExpr._from_c(c_stmt.Ist.Put.data), c_stmt.Ist.Put.offset)
 
     def typecheck(self, tyenv):
         return self.data.typecheck(tyenv)
@@ -186,9 +192,9 @@ class PutI(IRStmt):
     Write to a guest register, at a non-fixed offset in the guest state.
     """
 
-    __slots__ = ['descr', 'ix', 'data', 'bias']
+    __slots__ = ["descr", "ix", "data", "bias"]
 
-    tag = 'Ist_PutI'
+    tag = "Ist_PutI"
 
     def __init__(self, descr, ix, data, bias):
         self.descr = descr
@@ -196,22 +202,24 @@ class PutI(IRStmt):
         self.data = data
         self.bias = bias
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "PutI(%s)[%s,%d] = %s" % (self.descr, self.ix, self.bias, self.data)
 
     @staticmethod
     def _from_c(c_stmt):
-        return PutI(IRRegArray._from_c(c_stmt.Ist.PutI.details.descr),
-                    IRExpr._from_c(c_stmt.Ist.PutI.details.ix),
-                    IRExpr._from_c(c_stmt.Ist.PutI.details.data),
-                    c_stmt.Ist.PutI.details.bias)
+        return PutI(
+            IRRegArray._from_c(c_stmt.Ist.PutI.details.descr),
+            IRExpr._from_c(c_stmt.Ist.PutI.details.ix),
+            IRExpr._from_c(c_stmt.Ist.PutI.details.data),
+            c_stmt.Ist.PutI.details.bias,
+        )
 
     def typecheck(self, tyenv):
         dataty = self.data.typecheck(tyenv)
         if dataty is None:
             return False
         if dataty != self.descr.elemTy:
-            l.debug("Expression doesn't match RegArray type")
+            log.debug("Expression doesn't match RegArray type")
             return False
         return True
 
@@ -222,36 +230,35 @@ class WrTmp(IRStmt):
     will reject any block containing a temporary which is not assigned to exactly once.
     """
 
-    __slots__ = ['data', 'tmp']
+    __slots__ = ["data", "tmp"]
 
-    tag = 'Ist_WrTmp'
+    tag = "Ist_WrTmp"
 
-    def __init__(self, tmp: TmpVar, data: 'IRExpr'):
+    def __init__(self, tmp, data: "IRExpr"):
         self.tmp = tmp
         self.data = data
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         # Support for named register in string representation of expr.Get
 
         if arch is not None and tyenv is not None and isinstance(self.data, Get):
             reg_name = arch.translate_register_name(self.data.offset, self.data.result_size(tyenv) // 8)
 
         if reg_name is not None and isinstance(self.data, expr.Get):
-            return "t%d = %s" % (self.tmp, self.data.__str__(reg_name=reg_name))
+            return "t%d = %s" % (self.tmp, self.data.pp_str_with_name(reg_name))
         else:
             return "t%d = %s" % (self.tmp, self.data)
 
     @staticmethod
     def _from_c(c_stmt):
-        return WrTmp(c_stmt.Ist.WrTmp.tmp,
-                     IRExpr._from_c(c_stmt.Ist.WrTmp.data))
+        return WrTmp(c_stmt.Ist.WrTmp.tmp, IRExpr._from_c(c_stmt.Ist.WrTmp.data))
 
     def typecheck(self, tyenv):
         dataty = self.data.typecheck(tyenv)
         if dataty is None:
             return False
         if dataty != tyenv.lookup(self.tmp):
-            l.debug("Expression doesn't match tmp type")
+            log.debug("Expression doesn't match tmp type")
             return False
         return True
 
@@ -261,11 +268,11 @@ class Store(IRStmt):
     Write a value to memory..
     """
 
-    __slots__ = ['addr', 'data', 'end']
+    __slots__ = ["addr", "data", "end"]
 
-    tag = 'Ist_Store'
+    tag = "Ist_Store"
 
-    def __init__(self, addr: 'IRExpr', data: 'IRExpr', end: str):
+    def __init__(self, addr: "IRExpr", data: "IRExpr", end: str):
         self.addr = addr
         self.data = data
         self.end = end
@@ -274,14 +281,16 @@ class Store(IRStmt):
     def endness(self):
         return self.end
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
-        return "ST%s(%s) = %s" % (self.endness[-2:].lower(), self.addr, self.data)
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
+        return f"ST{self.endness[-2:].lower()}({self.addr}) = {self.data}"
 
     @staticmethod
     def _from_c(c_stmt):
-        return Store(IRExpr._from_c(c_stmt.Ist.Store.addr),
-                     IRExpr._from_c(c_stmt.Ist.Store.data),
-                     get_enum_from_int(c_stmt.Ist.Store.end))
+        return Store(
+            IRExpr._from_c(c_stmt.Ist.Store.addr),
+            IRExpr._from_c(c_stmt.Ist.Store.data),
+            get_enum_from_int(c_stmt.Ist.Store.end),
+        )
 
     def typecheck(self, tyenv):
         dataty = self.data.typecheck(tyenv)
@@ -291,10 +300,10 @@ class Store(IRStmt):
         if addrty is None:
             return False
         if addrty != tyenv.wordty:
-            l.debug("addr must be full word for arch")
+            log.debug("addr must be full word for arch")
             return False
-        if self.end not in ('Iend_LE', 'Iend_BE'):
-            l.debug("invalid endness enum")
+        if self.end not in ("Iend_LE", "Iend_BE"):
+            log.debug("invalid endness enum")
             return False
         return True
 
@@ -304,9 +313,9 @@ class CAS(IRStmt):
     an atomic compare-and-swap operation.
     """
 
-    __slots__ = ['addr', 'dataLo', 'dataHi', 'expdLo', 'expdHi', 'oldLo', 'oldHi', 'end']
+    __slots__ = ["addr", "dataLo", "dataHi", "expdLo", "expdHi", "oldLo", "oldHi", "end"]
 
-    tag = 'Ist_CAS'
+    tag = "Ist_CAS"
 
     def __init__(self, addr, dataLo, dataHi, expdLo, expdHi, oldLo, oldHi, end):
         self.addr = addr
@@ -322,43 +331,46 @@ class CAS(IRStmt):
     def endness(self):
         return self.end
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
-        return "t(%s,%s) = CAS%s(%s :: (%s,%s)->(%s,%s))" % (
-        self.oldLo, self.oldHi, self.end[-2:].lower(), self.addr, self.expdLo, self.expdHi, self.dataLo, self.dataHi)
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
+        return "t({},{}) = CAS{}({} :: ({},{})->({},{}))".format(
+            self.oldLo, self.oldHi, self.end[-2:].lower(), self.addr, self.expdLo, self.expdHi, self.dataLo, self.dataHi
+        )
 
     @staticmethod
     def _from_c(c_stmt):
-        return CAS(IRExpr._from_c(c_stmt.Ist.CAS.details.addr),
-                   IRExpr._from_c(c_stmt.Ist.CAS.details.dataLo),
-                   IRExpr._from_c(c_stmt.Ist.CAS.details.dataHi),
-                   IRExpr._from_c(c_stmt.Ist.CAS.details.expdLo),
-                   IRExpr._from_c(c_stmt.Ist.CAS.details.expdHi),
-                   c_stmt.Ist.CAS.details.oldLo,
-                   c_stmt.Ist.CAS.details.oldHi,
-                   get_enum_from_int(c_stmt.Ist.CAS.details.end))
+        return CAS(
+            IRExpr._from_c(c_stmt.Ist.CAS.details.addr),
+            IRExpr._from_c(c_stmt.Ist.CAS.details.dataLo),
+            IRExpr._from_c(c_stmt.Ist.CAS.details.dataHi),
+            IRExpr._from_c(c_stmt.Ist.CAS.details.expdLo),
+            IRExpr._from_c(c_stmt.Ist.CAS.details.expdHi),
+            c_stmt.Ist.CAS.details.oldLo,
+            c_stmt.Ist.CAS.details.oldHi,
+            get_enum_from_int(c_stmt.Ist.CAS.details.end),
+        )
 
     def typecheck(self, tyenv):
         addrty = self.addr.typecheck(tyenv)
         if addrty is None:
             return False
         if addrty != tyenv.wordty:
-            l.debug("addr must be full word for arch")
+            log.debug("addr must be full word for arch")
             return False
-        if self.end not in ('Iend_LE', 'Iend_BE'):
-            l.debug("invalid endness enum")
+        if self.end not in ("Iend_LE", "Iend_BE"):
+            log.debug("invalid endness enum")
             return False
 
         if self.oldHi == 0xFFFFFFFF:
             # single-element case
             if self.expdHi is not None or self.dataHi is not None:
-                l.debug("expdHi and dataHi must be None")
+                log.debug("expdHi and dataHi must be None")
                 return False
             expdLoTy = self.expdLo.typecheck(tyenv)
             dataLoTy = self.dataLo.typecheck(tyenv)
             if expdLoTy is None or dataLoTy is None:
                 return False
             if tyenv.lookup(self.oldLo) != expdLoTy or expdLoTy != dataLoTy:
-                l.debug("oldLo, expdL, dataLo must all have the same type")
+                log.debug("oldLo, expdL, dataLo must all have the same type")
                 return False
         else:
             # double-element case
@@ -368,10 +380,14 @@ class CAS(IRStmt):
             dataHiTy = self.dataHi.typecheck(tyenv)
             if expdLoTy is None or dataLoTy is None or expdHiTy is None or dataHiTy is None:
                 return False
-            if tyenv.lookup(self.oldLo) != expdLoTy or expdLoTy != dataLoTy or \
-               tyenv.lookup(self.oldHi) != expdHiTy or expdHiTy != dataHiTy or \
-               expdLoTy != expdHiTy:
-                l.debug("oldLo, expdLo, dataLo, oldHi, expdHi, dataHi must all have the same type")
+            if (
+                tyenv.lookup(self.oldLo) != expdLoTy
+                or expdLoTy != dataLoTy
+                or tyenv.lookup(self.oldHi) != expdHiTy
+                or expdHiTy != dataHiTy
+                or expdLoTy != expdHiTy
+            ):
+                log.debug("oldLo, expdLo, dataLo, oldHi, expdHi, dataHi must all have the same type")
                 return False
 
         return True
@@ -379,13 +395,13 @@ class CAS(IRStmt):
 
 class LLSC(IRStmt):
     """
-     Either Load-Linked or Store-Conditional, depending on STOREDATA. If STOREDATA is NULL then this is a Load-Linked,
-     else it is a Store-Conditional.
+    Either Load-Linked or Store-Conditional, depending on STOREDATA. If STOREDATA is NULL then this is a Load-Linked,
+    else it is a Store-Conditional.
     """
 
-    __slots__ = ['addr', 'storedata', 'result', 'end']
+    __slots__ = ["addr", "storedata", "result", "end"]
 
-    tag = 'Ist_LLSC'
+    tag = "Ist_LLSC"
 
     def __init__(self, addr, storedata, result, end):
         self.addr = addr
@@ -397,7 +413,7 @@ class LLSC(IRStmt):
     def endness(self):
         return self.end
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         if self.storedata is None:
             return "t%d = LD%s-Linked(%s)" % (self.result, self.end[-2:].lower(), self.addr)
         else:
@@ -405,46 +421,46 @@ class LLSC(IRStmt):
 
     @staticmethod
     def _from_c(c_stmt):
-        return LLSC(IRExpr._from_c(c_stmt.Ist.LLSC.addr),
-                    IRExpr._from_c(c_stmt.Ist.LLSC.storedata),
-                    c_stmt.Ist.LLSC.result,
-                    get_enum_from_int(c_stmt.Ist.LLSC.end))
+        return LLSC(
+            IRExpr._from_c(c_stmt.Ist.LLSC.addr),
+            IRExpr._from_c(c_stmt.Ist.LLSC.storedata),
+            c_stmt.Ist.LLSC.result,
+            get_enum_from_int(c_stmt.Ist.LLSC.end),
+        )
 
     def typecheck(self, tyenv):
         addrty = self.addr.typecheck(tyenv)
         if addrty is None:
             return False
         if addrty != tyenv.wordty:
-            l.debug("addr must be full word for arch")
+            log.debug("addr must be full word for arch")
             return False
-        if self.end not in ('Iend_LE', 'Iend_BE'):
-            l.debug("invalid endness enum")
+        if self.end not in ("Iend_LE", "Iend_BE"):
+            log.debug("invalid endness enum")
             return False
 
-
         if self.storedata is not None:
             # load-linked
             storety = self.storedata.typecheck(tyenv)
             if storety is None:
                 return False
 
-            if tyenv.lookup(self.result) != 'Ity_I1':
-                l.debug("result tmp must be Ity_I1")
+            if tyenv.lookup(self.result) != "Ity_I1":
+                log.debug("result tmp must be Ity_I1")
                 return False
 
         return True
 
 
 class MBE(IRStmt):
+    __slots__ = ["event"]
 
-    __slots__ = ['event']
-
-    tag = 'Ist_MBE'
+    tag = "Ist_MBE"
 
     def __init__(self, event):
         self.event = event
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "MBusEvent-" + self.event
 
     @staticmethod
@@ -453,10 +469,9 @@ class MBE(IRStmt):
 
 
 class Dirty(IRStmt):
+    __slots__ = ["cee", "guard", "args", "tmp", "mFx", "mAddr", "mSize", "nFxState"]
 
-    __slots__ = ['cee', 'guard', 'args', 'tmp', 'mFx', 'mAddr', 'mSize', 'nFxState']
-
-    tag = 'Ist_Dirty'
+    tag = "Ist_Dirty"
 
     def __init__(self, cee, guard, args, tmp, mFx, mAddr, mSize, nFxState):
         self.cee = cee
@@ -468,9 +483,10 @@ class Dirty(IRStmt):
         self.mSize = mSize
         self.nFxState = nFxState
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
-        return "t%s = DIRTY %s %s ::: %s(%s)" % (
-        self.tmp, self.guard, "TODO(effects)", self.cee, ','.join(str(a) for a in self.args))
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
+        return "t{} = DIRTY {} {} ::: {}({})".format(
+            self.tmp, self.guard, "TODO(effects)", self.cee, ",".join(str(a) for a in self.args)
+        )
 
     @property
     def child_expressions(self):
@@ -490,14 +506,16 @@ class Dirty(IRStmt):
 
             args.append(IRExpr._from_c(a))
 
-        return Dirty(IRCallee._from_c(c_stmt.Ist.Dirty.details.cee),
-                     IRExpr._from_c(c_stmt.Ist.Dirty.details.guard),
-                     tuple(args),
-                     c_stmt.Ist.Dirty.details.tmp,
-                     get_enum_from_int(c_stmt.Ist.Dirty.details.mFx),
-                     IRExpr._from_c(c_stmt.Ist.Dirty.details.mAddr),
-                     c_stmt.Ist.Dirty.details.mSize,
-                     c_stmt.Ist.Dirty.details.nFxState)
+        return Dirty(
+            IRCallee._from_c(c_stmt.Ist.Dirty.details.cee),
+            IRExpr._from_c(c_stmt.Ist.Dirty.details.guard),
+            tuple(args),
+            c_stmt.Ist.Dirty.details.tmp,
+            get_enum_from_int(c_stmt.Ist.Dirty.details.mFx),
+            IRExpr._from_c(c_stmt.Ist.Dirty.details.mAddr),
+            c_stmt.Ist.Dirty.details.mSize,
+            c_stmt.Ist.Dirty.details.nFxState,
+        )
 
 
 class Exit(IRStmt):
@@ -505,9 +523,9 @@ class Exit(IRStmt):
     A conditional exit from the middle of an IRSB.
     """
 
-    __slots__ = ['guard', 'dst', 'offsIP', 'jk']
+    __slots__ = ["guard", "dst", "offsIP", "jk"]
 
-    tag = 'Ist_Exit'
+    tag = "Ist_Exit"
 
     def __init__(self, guard, dst, jk, offsIP):
         self.guard = guard
@@ -519,36 +537,37 @@ class Exit(IRStmt):
     def jumpkind(self):
         return self.jk
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
-
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         if arch is not None and tyenv is not None:
             reg_name = arch.translate_register_name(self.offsIP, arch.bits // 8)
 
         if reg_name is None:
             return "if (%s) { PUT(offset=%d) = %#x; %s }" % (self.guard, self.offsIP, self.dst.value, self.jumpkind)
         else:
-            return "if (%s) { PUT(%s) = %#x; %s }" % (self.guard, reg_name, self.dst.value, self.jumpkind)
+            return f"if ({self.guard}) {{ PUT({reg_name}) = {self.dst.value:#x}; {self.jumpkind} }}"
 
     @property
     def child_expressions(self):
-        return [self.guard, self.dst] + self.guard.child_expressions
+        return [self.guard] + self.guard.child_expressions + [Const(self.dst)]
 
     @staticmethod
     def _from_c(c_stmt):
-        return Exit(IRExpr._from_c(c_stmt.Ist.Exit.guard),
-                    IRConst._from_c(c_stmt.Ist.Exit.dst),
-                    get_enum_from_int(c_stmt.Ist.Exit.jk),
-                    c_stmt.Ist.Exit.offsIP)
+        return Exit(
+            IRExpr._from_c(c_stmt.Ist.Exit.guard),
+            IRConst._from_c(c_stmt.Ist.Exit.dst),
+            get_enum_from_int(c_stmt.Ist.Exit.jk),
+            c_stmt.Ist.Exit.offsIP,
+        )
 
     def typecheck(self, tyenv):
         if not self.jk.startswith("Ijk_"):
-            l.debug("Jumpkind is not a jumpkind enum")
+            log.debug("Jumpkind is not a jumpkind enum")
             return False
         guardty = self.guard.typecheck(tyenv)
         if guardty is None:
             return False
-        if guardty != 'Ity_I1':
-            l.debug("guard must be Ity_I1")
+        if guardty != "Ity_I1":
+            log.debug("guard must be Ity_I1")
             return False
         return True
 
@@ -558,9 +577,9 @@ class LoadG(IRStmt):
     A guarded load.
     """
 
-    __slots__ = ['addr', 'alt', 'guard', 'dst', 'cvt', 'end', 'cvt_types']
+    __slots__ = ["addr", "alt", "guard", "dst", "cvt", "end", "cvt_types"]
 
-    tag = 'Ist_LoadG'
+    tag = "Ist_LoadG"
 
     def __init__(self, end, cvt, dst, addr, alt, guard):
         self.addr = addr
@@ -570,39 +589,47 @@ class LoadG(IRStmt):
         self.cvt = cvt
         self.end = end
 
-        type_in = ffi.new('IRType *') # TODO separate this from the pyvex C implementation
-        type_out = ffi.new('IRType *')
+        type_in = ffi.new("IRType *")  # TODO separate this from the pyvex C implementation
+        type_out = ffi.new("IRType *")
         pvc.typeOfIRLoadGOp(get_int_from_enum(self.cvt), type_out, type_in)
-        type_in = ffi.cast('int *', type_in)[0]
-        type_out = ffi.cast('int *', type_out)[0]
+        type_in = ffi.cast("int *", type_in)[0]
+        type_out = ffi.cast("int *", type_out)[0]
         self.cvt_types = (get_enum_from_int(type_in), get_enum_from_int(type_out))
 
     @property
     def endness(self):
         return self.end
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
         return "t%d = if (%s) %s(LD%s(%s)) else %s" % (
-        self.dst, self.guard, self.cvt, self.end[-2:].lower(), self.addr, self.alt)
+            self.dst,
+            self.guard,
+            self.cvt,
+            self.end[-2:].lower(),
+            self.addr,
+            self.alt,
+        )
 
     @staticmethod
     def _from_c(c_stmt):
-        return LoadG(get_enum_from_int(c_stmt.Ist.LoadG.details.end),
-                     get_enum_from_int(c_stmt.Ist.LoadG.details.cvt),
-                     c_stmt.Ist.LoadG.details.dst,
-                     IRExpr._from_c(c_stmt.Ist.LoadG.details.addr),
-                     IRExpr._from_c(c_stmt.Ist.LoadG.details.alt),
-                     IRExpr._from_c(c_stmt.Ist.LoadG.details.guard))
+        return LoadG(
+            get_enum_from_int(c_stmt.Ist.LoadG.details.end),
+            get_enum_from_int(c_stmt.Ist.LoadG.details.cvt),
+            c_stmt.Ist.LoadG.details.dst,
+            IRExpr._from_c(c_stmt.Ist.LoadG.details.addr),
+            IRExpr._from_c(c_stmt.Ist.LoadG.details.alt),
+            IRExpr._from_c(c_stmt.Ist.LoadG.details.guard),
+        )
 
     def typecheck(self, tyenv):
         addrty = self.addr.typecheck(tyenv)
         if addrty is None:
             return False
         if addrty != tyenv.wordty:
-            l.debug("addr must be full word for arch")
+            log.debug("addr must be full word for arch")
             return False
-        if self.end not in ('Iend_LE', 'Iend_BE'):
-            l.debug("invalid endness enum")
+        if self.end not in ("Iend_LE", "Iend_BE"):
+            log.debug("invalid endness enum")
             return False
 
         dstty = tyenv.lookup(self.dst)
@@ -611,14 +638,14 @@ class LoadG(IRStmt):
 
         if guardty is None or altty is None:
             return False
-        if dstty != 'Ity_I32' or altty != 'Ity_I32':
-            l.debug('dst and alt must be Ity_I32')
+        if dstty != "Ity_I32" or altty != "Ity_I32":
+            log.debug("dst and alt must be Ity_I32")
             return False
-        if guardty != 'Ity_I1':
-            l.debug('guard must be Ity_I1')
+        if guardty != "Ity_I1":
+            log.debug("guard must be Ity_I1")
             return False
-        if not self.cvt.startswith('ILGop_'):
-            l.debug("Invalid cvt enum")
+        if not self.cvt.startswith("ILGop_"):
+            log.debug("Invalid cvt enum")
             return False
         return True
 
@@ -628,9 +655,9 @@ class StoreG(IRStmt):
     A guarded store.
     """
 
-    __slots__ = ['addr', 'data', 'guard', 'end']
+    __slots__ = ["addr", "data", "guard", "end"]
 
-    tag = 'Ist_StoreG'
+    tag = "Ist_StoreG"
 
     def __init__(self, end, addr, data, guard):
         self.addr = addr
@@ -642,25 +669,27 @@ class StoreG(IRStmt):
     def endness(self):
         return self.end
 
-    def __str__(self, reg_name=None, arch=None, tyenv=None):
-        return "if (%s) ST%s(%s) = %s" % (self.guard, self.end[-2:].lower(), self.addr, self.data)
+    def pp_str(self, reg_name=None, arch=None, tyenv=None):
+        return f"if ({self.guard}) ST{self.end[-2:].lower()}({self.addr}) = {self.data}"
 
     @staticmethod
     def _from_c(c_stmt):
-        return StoreG(get_enum_from_int(c_stmt.Ist.StoreG.details.end),
-                      IRExpr._from_c(c_stmt.Ist.StoreG.details.addr),
-                      IRExpr._from_c(c_stmt.Ist.StoreG.details.data),
-                      IRExpr._from_c(c_stmt.Ist.StoreG.details.guard))
+        return StoreG(
+            get_enum_from_int(c_stmt.Ist.StoreG.details.end),
+            IRExpr._from_c(c_stmt.Ist.StoreG.details.addr),
+            IRExpr._from_c(c_stmt.Ist.StoreG.details.data),
+            IRExpr._from_c(c_stmt.Ist.StoreG.details.guard),
+        )
 
     def typecheck(self, tyenv):
         addrty = self.addr.typecheck(tyenv)
         if addrty is None:
             return False
         if addrty != tyenv.wordty:
-            l.debug("addr must be full word for arch")
+            log.debug("addr must be full word for arch")
             return False
-        if self.end not in ('Iend_LE', 'Iend_BE'):
-            l.debug("invalid endness enum")
+        if self.end not in ("Iend_LE", "Iend_BE"):
+            log.debug("invalid endness enum")
             return False
 
         guardty = self.guard.typecheck(tyenv)
@@ -668,8 +697,8 @@ class StoreG(IRStmt):
 
         if guardty is None or dataty is None:
             return False
-        if guardty != 'Ity_I1':
-            l.debug('guard must be Ity_I1')
+        if guardty != "Ity_I1":
+            log.debug("guard must be Ity_I1")
             return False
         return True
 
@@ -678,8 +707,8 @@ _globals = globals().copy()
 #
 # Mapping from tag strings/enums to IRStmt classes
 #
-tag_to_stmt_mapping = { }
-enum_to_stmt_mapping = { }
+tag_to_stmt_mapping = {}
+enum_to_stmt_mapping = {}
 tag_count = 0
 cls = None
 for cls in _globals.values():
@@ -690,22 +719,16 @@ for cls in _globals.values():
         tag_count += 1
 del cls
 
+
 def tag_to_stmt_class(tag):
     try:
         return tag_to_stmt_mapping[tag]
     except KeyError:
-        raise KeyError('No statement class for tag %s.' % tag)
+        raise KeyError("No statement class for tag %s." % tag)
 
 
 def enum_to_stmt_class(tag_enum):
     try:
         return enum_to_stmt_mapping[tag_enum]
     except KeyError:
-        raise KeyError('No statement class for tag %s.' % get_enum_from_int((tag_enum)))
-
-
-from .expr import IRExpr, Get
-from .const import IRConst
-from .enums import IRRegArray, IRCallee
-from .errors import PyVEXError
-from . import ffi, pvc, expr
+        raise KeyError("No statement class for tag %s." % get_enum_from_int(tag_enum))
